
#include <sys/stat.h>
#include <bits/posix1_lim.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>


int unlinkat(int dirfd, const char *pathname, int flags)
{
	int (*real_unlinkat)(int dirfd, const char *pathname, int flags) = dlsym(RTLD_NEXT, "unlinkat");
	int retval;

	char *recyclers;
	char *ptr, *tmp;
	char recycler[_POSIX_PATH_MAX];
	char recycled[_POSIX_PATH_MAX];
	char atdir[_POSIX_PATH_MAX];
	char fqpath[_POSIX_PATH_MAX];
	char mp_path[_POSIX_PATH_MAX];
	struct stat st1;
	struct stat st2;
	char *txt_unset = "unset RECYCLER environment if you do not care about Recycle bin";

	recyclers = getenv("RECYCLER");
	if(recyclers != NULL)
	{
		retval = fstatat(dirfd, pathname, &st1, flags & ~AT_REMOVEDIR /* fstatat hates AT_REMOVEDIR */ | AT_SYMLINK_NOFOLLOW /* we are about to move/unlink the symlink itself */);
		if(retval != 0)
		{
			warn("%s", pathname);
		}
		else
		{
			while(1)
			{
				ptr = strchrnul(recyclers, ':');
				strncpy(recycler, recyclers, ptr - recyclers);
				recycler[ptr - recyclers] = '\0';

				if(strncmp(recycler, "~/", 2) == 0)
				{
					tmp = getenv("HOME");
					if(tmp == NULL)
					{
						warnx("Could not resolve: %s", recycler);
						recycler[0] = '\0';
					}
					else
					{
						snprintf(recycler, strlen(tmp) + 1 + (ptr - recyclers - 2) + 1, "%s/%s\000", tmp, recyclers + 2);
					}
				}
				
				if(strlen(recycler) > 0)
				{
					#ifdef DEBUG
					warnx("RECYCLER %s", recycler);
					#endif
					
					retval = stat(recycler, &st2);
					if(retval != 0)
					{
						warn("stat: %s", recycler);
					}
					else
					{
						if(st2.st_dev == st1.st_dev)
						{
							/* Recycler and the file to be removed are on the same fs */
							/* ptr no longer points to the next RECYCLER */
							
							if(pathname[0] == '/')
							{
								/* Absolute pathname given */
								strcpy(fqpath, pathname);
							}
							else
							{
								/* Relative pathname given */
								if(dirfd == AT_FDCWD)
								{
									if(getcwd(atdir, _POSIX_PATH_MAX) == NULL)
									{
										warn("getcwd");
										retval = -1;
										break;
									}
								}
								else
								{
									sprintf(fqpath, "/proc/self/fd/%d", dirfd);
									retval = readlink(fqpath, atdir, _POSIX_PATH_MAX);
									if(retval == -1)
									{
										warn("readlink: %s", fqpath);
										break;
									}
									atdir[retval] = '\0';
								}
								
								sprintf(fqpath, "%s/%s", atdir, pathname);
							}
							
							strcpy(atdir, fqpath);
							tmp = (char*)dirname(atdir);  /* atdir got changed */
							ptr = realpath(tmp, NULL);
							if(ptr == NULL)
							{
								warn("realpath: %s", tmp);
								free(ptr);
								break;
							}
							free(ptr);
							/* ptr now points to the resolved path name of fqpath's parent directory */
							strcpy(atdir, fqpath);
							tmp = (char*)basename(atdir);  /* atdir got changed */
							sprintf(fqpath, "%s/%s", ptr, tmp);
							
							retval = 0;
							strcpy(mp_path, fqpath);
							while(1)
							{
								/* Looking for filesystem's root */
								strcpy(atdir, mp_path);
								tmp = (char *)dirname(atdir);  /* dirname will change atdir! */
								if(strcmp(tmp, "//") == 0)
								{
									tmp[1] = '\0';
								}
								retval = stat(tmp, &st2);
								if(retval != 0)
								{
									warn("stat: %s", tmp);
									break;
								}

								if(strcmp(mp_path, "/") == 0 && strcmp(tmp, "/") == 0)
								{
									/* fqpath is on the root fs */
									break;
								}
								if(st1.st_dev != st2.st_dev)
								{
									/* mp_path is the mountpoint */
									break;
								}
								strcpy(mp_path, tmp);
							}
							if(retval != 0) break;
							
							sprintf(recycled, "%s/%s", recycler, fqpath + strlen(mp_path));
							#ifdef DEBUG
							warnx("Filesystem root: %s", mp_path);
							warnx("Recycled file: %s", recycled);
							#endif
							
							/* Copy directory path onto Recycler */
							strcpy(atdir, recycled);
							if(S_ISDIR(st1.st_mode))
							{
								/* Clone the file (actually the directory) to be removed itself */
								tmp = atdir;
								ptr = fqpath;
							}
							else
							{
								/* Clone the file's parent directory  */
								tmp = (char*)dirname(atdir);
								ptr = (char*)dirname(fqpath);
							}
							retval = recyclix_clonepath_recursive(tmp, ptr);
							if(retval != 0)
							{
								warn("mkdir: %s", tmp);
								break;
							}
							
							if(S_ISDIR(st1.st_mode))
							{
								goto real_unlink;
							}
							else
							{
								retval = renameat(dirfd, pathname, 0, recycled);
								if(retval != 0)
								{
									warn("renameat");
									warnx("Failed to drop file to the Recycler, %s.", txt_unset);
								}
							}
							
							/* ptr is unusable to continue */
							break;
						}
						#ifdef DEBUG
						else
						{
							/* Cross-device condition */
							warnx("xdev: %03u,%03u (looking for %03u,%03u)", major(st2.st_dev), minor(st2.st_dev), major(st1.st_dev), minor(st1.st_dev));
						}
						#endif
						/* Looking for next Recycler ... */
					}
				}
				
				if(ptr[0] == '\0')
				{
					#ifdef DEBUG
					warnx("There is no valid Recycler, removing file.");
					#endif
					goto real_unlink;
					break;
				}
				else
				{
					recyclers = ptr+1;
				}
			}
		}
	}
	else
	{
		real_unlink:
		retval = real_unlinkat(dirfd, pathname, flags);
	}

	return retval;
}

int recyclix_clonepath_recursive(char * p_path, char * p_donor)
{
	int (*real_unlink)(const char *pathname) = dlsym(RTLD_NEXT, "unlink");
	int retval;
	char path[_POSIX_PATH_MAX];
	char donor[_POSIX_PATH_MAX];
	char * parent;
	char * dparent;
	struct stat st;
	struct stat donor_st;
	
	#ifdef DEBUG
	warnx("%s: %s", __func__, p_path);
	#endif

	/* save stats of directory being cloned */	
	retval = stat(p_donor, &donor_st);
	if(retval == -1)
	{
		warn("stat: %s", p_donor);
	}
	else
	{
		try_mkdir:
		retval = mkdir(p_path, donor_st.st_mode);
		if(retval == -1)
		{
			if(errno == ENOENT || errno == ENOTDIR)
			{
				strcpy(path, p_path);
				strcpy(donor, p_donor);
				parent = (char*)dirname(path);
				dparent = (char*)dirname(donor);
				
				retval = real_unlink(parent);
				if(retval == 0)
				{
					goto try_mkdir;
				}
				else if(errno == ENOENT)
				{
					retval = recyclix_clonepath_recursive(parent, dparent);
					if(retval == 0)
					{
						goto try_mkdir;
					}
				}
				else
				{
					warn("unlink: %s", parent);
				}
			}
			else if(errno == EEXIST)
			{
				retval = stat(p_path, &st);
				if(retval == -1)
				{
					warn("stat: %s", p_path);
				}
				else
				{
					if(S_ISDIR(st.st_mode))
					{
						chmod(p_path, donor_st.st_mode);
						retval = 0;
					}
					else
					{
						retval = real_unlink(p_path);
						if(retval == 0)
						{
							goto try_mkdir;
						}
						else
						{
							warn("unlink: %s", p_path);
						}
					}
				}
			}
			else
			{	
				warn("mkdir: %s", p_path);
			}
		}
	}
	if(retval == 0)
	{
		chown(p_path, donor_st.st_uid, donor_st.st_gid);
	}
	return retval;
}

int unlink(const char *pathname)
{
	return unlinkat(AT_FDCWD, pathname, 0);
}

